#######################################################################
#
# ABOUT THIS FILE:  This file is intended to make work with tux parts a
#                   bit simpler by offering a possibility to automatically
#                   render all parts of the TUX GUN MOTION at once as
#                   well as only rendering the image series for one 
#                   particular part of the Tux.
#                   I hope it really does the trick. 
#
# STEPS:  Several things must be done before this script will really do some work.
#
# 0. Blender must be started from FreeDroid/graphics/blender_sources directory.
#
# 1. Coppy (part of FreedroidRPG) must be compiled in FreeDroid/croppy/ dir.
#    (The path where to find croppy must be entered below in variable 'croppy_path')
#
# 2. /tmp/ directory must be writable for common user.  This is because
#    blender output will go directly to that dir before it is properly
#    cropped by the croppy tool and then placed in the proper subdir of
#    the FreeDroid directory tree.
# 
# 3. No matter if rendering every part of the Tux or just one part of the
#    Tux, the visible layers for the rendering should be only the first 
#    layer.
#    (didn't find a means to set the visible layers from inside the script
#     and question in elysiun python forum didn't turn up anything either,
#     ...)
#
#--------------------
#
# RENDERING ONE PART OF THE TUX REQUIRES THE FOLLOWING: 
#
#  * the variable 'body_part_string' should be adapted to the part desired.
#    (in general that means uncommenting the right 'part_string' line and
#    commenting out all other part_string definition lines down there.
#
#  * The variable 'do_ultimate_rendering' should be set to 0
#
#  * Set the last frame of the animation to something >= 16 * 36, e.g.
#    you could set the last frame to 580.  That should be safe.  If you
#    use more that that, it should just redo some work already done, but
#    no harm will come from it.
#
# RENDERING ALL PARTS OF THE TUX IN ONE BIG GO REQUIRES THE FOLLOWING:
#
#  * The variable 'do_ultimate_rendering' should be set to 1
#
#  * Set the last frame of the animation to something >= 13 * 16 * 36, e.g.
#    you could set the last frame to 5800.  That should be safe.  If you
#    use more that that, it should just redo some work already done, but
#    no harm will come from it.
#    The factor 13 in this calculation comes from the fact, that there
#    are 13 different body parts possible.  Once we add more, we should
#    add also to the last frame number.
#
#--------------------
#
# COMMENTS:
#
# an occasional error message 'cannot stat ...0000.png' is completely ok.  
# The script should still work as intended.
#
#----------------------------------------------------------------------


######################################################################
######################### PREPARE EVERYTHING #########################
######################################################################

######################
# Let's start with the preparations:  Importing modules...
#
import Blender
import os
import sys
from math import *

######################
# Define some constants for later computation
#
pi2 = pi * 2 
number_of_tux_directions = 16.0
rot_angle = pi2 / number_of_tux_directions
our_cam = Blender.Object.Get("RotoCamera")

######################
# Define some file pathes...
#
croppy_path = "./../../croppy/croppy"
mv_path = "mv"
rm_path = "rm"
source_directory = "/tmp/"
target_directory = "./../tux_motion_parts/gun_motion/"


######################
# The length of the full motion cycle is the last frame that still
# contains some keyframes of the animation.  It need not ever be changed
# unless we invent a completely new kind of Tux motion, like falling down
# and standing back up again or something, which would need to be appended
# after the running part of the motion.
#
full_motion_cycle = 36

######################
# Here comes the ultimate rendering switch!  Attention!
# THIS SWITCH WILL START THE GENERATION OF EVERY TUX GUN MOTION IMAGE!
#
# BEST WILL BE TO LET THIS RUN OVERNIGHT, PROVIDED YOUR MACHINE IS FIRE-PROOF
# AND WILL NOT OVERHEAT AND BURN DOWN THE HOUSE...
#
# Set this switch to 1 for ultimate rendering and to something else for normal
# operation.
#
do_ultimate_rendering = 1 

######################
# Unless we intend to use this script for generation of everything from the
# command line (which is now also possible) we should select some part that
# will be the target of the whole rendering operation.
#
# body_part_string = "head"
# body_part_string = "torso"
# body_part_string = "helm1"
# body_part_string = "armour1"
# body_part_string = "robe"
# body_part_string = "shieldarm"
# body_part_string = "weaponarm"
# body_part_string = "gun1"
# body_part_string = "feet"
body_part_string = "boots1"

######################
# We now get a hold upon every object in the file, so that
# we can later arrange our scene with only the parts we want
# to see at any given moment and all from the script alone
#
our_gun               = Blender.Object.Get("TheGun")
our_head              = Blender.Object.Get("TheHead")
our_legs              = Blender.Object.Get("TheLegs")
our_left_foot         = Blender.Object.Get("TheLeftFoot")
our_right_foot        = Blender.Object.Get("TheRightFoot")
our_torso             = Blender.Object.Get("TheTorso")
our_left_arm          = Blender.Object.Get("TheLeftArm")
our_left_glove        = Blender.Object.Get("TheLeftGlove")
our_right_arm         = Blender.Object.Get("TheRightArm")
our_yellow_helmet     = Blender.Object.Get("TheYellowHelmet")
our_left_boot         = Blender.Object.Get("TheLeftBoot")
our_right_boot        = Blender.Object.Get("TheRightBoot")

our_yellow_armour     = Blender.Object.Get("TheYellowArmour")
our_robe              = Blender.Object.Get("TheRobe")

our_origin_spot       = Blender.Object.Get("TheOriginSpot")

######################
# It might become nescessary, that we remove all the old parts,
# since those could interfere with out current part rendering, so
# we make a function to take care of that and remove everything back
# to normal.
#
def move_everything_to_home_layer():
	our_gun.Layer                = 1
	our_head.Layer               = 2 
	our_legs.Layer               = 2 * 2 
	our_left_foot.Layer          = 2 * 2
	our_right_foot.Layer         = 2 * 2 
	our_torso.Layer              = 2 * 2 * 2
	our_left_arm.Layer           = 2 * 2 * 2 * 2
	our_left_glove.Layer         = 2 * 2 * 2 * 2
	our_right_arm.Layer          = 2 * 2 * 2 * 2 * 2
	our_yellow_helmet.Layer      = 2*2*2*2 * 2*2*2*2*2 * 2*2
	our_left_boot.Layer          = 2*2*2*2 * 2*2*2*2*2 * 2*2*2
	our_right_boot.Layer         = 2*2*2*2 * 2*2*2*2*2 * 2*2*2
	our_yellow_armour.Layer      = 2*2*2*2 * 2*2*2*2*2 * 2*2*2*2
	our_robe.Layer               = 2*2*2*2 * 2*2*2*2*2 * 2*2*2*2*2 * 2*2*2*2
	our_origin_spot.Layer        = 2*2*2*2 * 2*2*2*2*2 * 2*2*2*2*2 * 2*2*2*2*2	
	Blender.Redraw()

######################
# The following function should prepare the current scene in the sense,
# that all nescessary layer operation are performed, like moving unused
# Tux parts to different layers and moving the needed Tux parts to the
# visible layer, which will be layer 1 for reasons of long-time tradition.
#
def prepare_scene_for_tux_part():
	global stop_file_generation_after_image_nr
	global subdirectory_string
	move_everything_to_home_layer()
	our_gun.Layer         = 2*2*2*2 * 2*2*2*2*2 * 2*2*2*2*2 * 2*2*2*2*2
	print "NOW STARTING TO PREPARE SCENE FOR THIS TUX PART."
	if ( body_part_string == "gun1" ) :
		our_gun.Layer          = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 16
	elif ( body_part_string == "head" ) :
		our_head.Layer               = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 16
	elif ( body_part_string == "helm1" ) :
		our_head.Layer      = 1
		our_yellow_helmet.Layer      = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 16
	elif ( body_part_string == "shieldarm" ) :
		our_left_arm.Layer               = 1
		our_left_glove.Layer               = 1
		subdirectory_string = "shield_items/"
		stop_file_generation_after_image_nr = 36
	elif ( body_part_string == "weaponarm" ) :
		our_right_arm.Layer      = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 36
	elif ( body_part_string == "boots1" ) :
		our_legs.Layer      = 1
		our_left_boot.Layer      = 1
		our_right_boot.Layer      = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 36
	elif ( body_part_string == "feet" ) :
		our_legs.Layer      = 1
		our_left_foot.Layer      = 1
		our_right_foot.Layer      = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 36
	elif ( body_part_string == "torso" ) :
		our_torso.Layer      = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 36
	elif ( body_part_string == "armour1" ) :
		our_torso.Layer      = 1
		our_yellow_armour.Layer      = 1
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 36
	elif ( body_part_string == "robe" ) :
		our_torso.Layer      = 1
		our_robe.Layer      = 1
		Blender.Redraw()
		subdirectory_string = ""
		stop_file_generation_after_image_nr = 36
	else:
		print "******************************"
		print "*** F A T A L   E R R O R   O C C U R E D   ! ! !"
		print "***"
		print "*** --> Part string was not identified.  breaking off..."
		print "***"
		print " "
		######################
		# The bad thing is:  sys.exit only quits this script for this frame,
		# but never does anything else, so next frame, we'll be back here :)
		#
		# os.abort()
		# sys.exit()
		print "*** THIS SHOULD MUST NOT EVER BE VISIBLE"
		print "***"
		Blender.Redraw()
		subdirectory_string = "NOWHERE_TO_BE_FOUND"
		print "The body part name encountered was: " , body_part_string
		stop_file_generation_after_image_nr = 36

######################################################################
################## FIND OUT THE OUTPUT DIRECTION #####################
######################################################################

######################
# Now we can start to identify the current direction
# by taking a closer look at the current camera position.
#
output_direction = 0 
start_x_pos = our_cam.LocX
start_y_pos = our_cam.LocY
#
# Now we need to be careful to avoid division by zero...
#
if our_cam.LocX == 0:
	our_arctan = atan ( our_cam.LocY / 0.0001 )
else:
	our_arctan = atan ( our_cam.LocY / our_cam.LocX )

def get_output_direction_16():
	global output_direction
	if abs( our_arctan - 0 * pi/4 ) < pi2/32:
		if start_x_pos < 0 :
			output_direction = 4
		else:
			output_direction = 12
	elif abs( our_arctan - 1 * pi/8 ) < pi2/32:
		if start_x_pos < 0 :
			output_direction = 3
		else:
			output_direction = 11
	elif abs( our_arctan + 1 * pi/8 ) < pi2/32:
		if start_x_pos < 0 :
			output_direction = 5
		else:
			output_direction = 13
	elif abs( our_arctan - 1 * pi/4 ) < pi2/32:
		if start_x_pos < 0 :
			output_direction = 2
		else:
			output_direction = 10
	elif abs( our_arctan + 1 * pi/4 ) < pi2/32:
		if start_y_pos < 0 :
			output_direction = 14
		else:
			output_direction = 6
	elif abs( our_arctan + 3 * pi/8 ) < pi2/32:
		if start_y_pos < 0 :
			output_direction = 15
		else:
			output_direction = 7
	elif abs( our_arctan - 3 * pi/8 ) < pi2/32:
		if start_y_pos < 0 :
			output_direction = 1
		else:
			output_direction = 9
	elif abs( our_arctan - 2 * pi/4 ) < pi2/32:
		if start_y_pos < 0 :
			output_direction = 0
		else:
			output_direction = 8
	elif abs( our_arctan + 2 * pi/4 ) < pi2/32:
		if start_y_pos < 0 :
			output_direction = 0
		else:
			output_direction = 8
	else:
		print "ERROR:  UNABLE TO DETERMINE DIRECTION!"
		Exit()

if number_of_tux_directions == 16:
	get_output_direction_16()
elif number_of_tux_directions == 8:
	get_output_direction_8()
else:
	print "ERROR:  illegal number of directions given."
	Exit()
print "BLENDER (MAIN): Final output direction: " , output_direction

######################################################################
##################### CAMERA ROTATION FUNCTION #######################
######################################################################

#######################
# Now we define our function that will rotate the camera
# by a fraction of a full turn
#
def do_cam_rot():
	r = our_cam.LocX * our_cam.LocX + our_cam.LocY * our_cam.LocY
	# print "Radius square: " , r
	r = sqrt ( r )
	# print "Radius used (without Z component):" , r
	######################
	# First we print out the starting position and rotation of the camera, just
	# so that we can check everything is all right, if we want to...
	#
	# print "Starting with coordinates: " , our_cam.loc
	# print "Starting with rotation: " , our_cam.rot
	######################
	# Now it's time to start doing the rotation.  Hope everything will go well...
	#
	X_new = 1.0 * ( cos ( rot_angle ) * our_cam.LocX + sin ( rot_angle ) * our_cam.LocY )
	Y_new = 1.0 * (-sin ( rot_angle ) * our_cam.LocX + cos ( rot_angle ) * our_cam.LocY )
	
	our_cam.LocX = X_new
	our_cam.LocY = Y_new
	our_cam.RotZ = our_cam.RotZ - rot_angle
	Blender.Redraw()

######################################################################
###################### RENAME AND CROPPY FUNCTION ####################
######################################################################

########################
# Now we check if maybe the last frame of the animation
# has been reached.  In this case we will do a rotation 
# of the camera.  If not we will start a croppy operation
# on the current output file.
#
time = Blender.Get("curframe") - 1

def do_rename_and_croppy():
	global output_direction
	global subdirectory_string
	global body_part_string
	global target_directory
	#
	#--------------------
	# Now we find out the number code of the source image file
	# on the disk, that has been created by the blender by now.
	# With this we can fix the source file name now.
	#
	if time < 10 :
		number_code = "0" + "0" + "0" + str(time)
	elif time < 100 :
		number_code = "0" + "0" + str(time)
	elif time < 1000 :
		number_code = "0" + str(time)
	else :
		number_code = str(time)
	source_filename = source_directory + "iso_gun_tmpfile_" + number_code + ".png"
	print "BLENDER (do_rename_and_croppy): Source file name: " , source_filename
	#
	#--------------------
	# Now we find out the end of the target file name, i.e. the last number code.
	# We introduce some redundancy to be safe against some last or first image in
	# the series getting lost somehow.  That makes up 3% of extra redundant computation
	# but that should be a tolerable price.
	#
	target_time = ( ( time % ( full_motion_cycle - 0 ) ) % ( full_motion_cycle - 1 ) ) + 1
	if target_time < 10 :
		target_number_code = "0" + "0" + "0" + str(target_time)
	else :
		target_number_code = "0" + "0" + str(target_time)
	#
	#--------------------
	# Now we fix the output number code.  The output number code can be derived from
	# the current camera position.
	#
	get_output_direction_16()
	if output_direction < 10 :
		direction_code = "0" + str ( output_direction )
	else:
		direction_code = str ( output_direction )
	#
	#--------------------
	# Now we should have all the information we need to assemble the final target
	# file name.
	#
	target_filename = target_directory + subdirectory_string
	target_filename = target_filename + "iso_" + body_part_string + "_" + direction_code
	target_filename = target_filename + "_" + target_number_code + ".png"
	print "BLENDER (do_rename_and_croppy): target file name would be: " , target_filename
	#
	#--------------------
	# Now we can start the croppy operation.
	#
	if ( ( target_time != 0 ) & ( target_time <= stop_file_generation_after_image_nr ) ) :
		# os.system ( "pwd" )
		command_line = mv_path + " " + source_filename + " " + target_filename
		print "BLENDER (do_rename_and_croppy): invoking 1st command line: " , command_line
		os.system ( command_line )
		# ----
		parameter = " -i " + target_filename
		print "BLENDER (do_rename_and_croppy): invoking 2nd command line: " , croppy_path , parameter
		os.system ( croppy_path + parameter )
	else :
		print "BLENDER (do_rename_and_croppy): This image will not be needed by the engine "
		print "                                and therefore we delete it."
		command_line = rm_path + " " + source_filename
		os.system ( command_line )
		# ----
	print " "

######################################################################
###################### ACTIVATION OF A BODY PART #####################
######################################################################

########################
# For a render-everything run we need to activate one body part.
# The part to be activated depends on the current time.
#

def set_current_part_string() :
	global body_part_string
	global time
	part_code = int ( ( int ( time + 0 ) ) / ( int ( ( full_motion_cycle ) * number_of_tux_directions ) ) )
	if ( part_code == (-1) ) :
		body_part_string = "head"
		# use images 0-576 for this
	elif ( part_code == 0 ) :
		body_part_string = "head"
		# use images 0-576 for this
	elif ( part_code == 1 ) :
		body_part_string = "torso"
		# use images 576-1152 for this
	elif ( part_code == 2 ) :
		body_part_string = "helm1"
		# use images 1152-1728 for this
	elif ( part_code == 3 ) :
		body_part_string = "armour1"
		# use images 1728-2304 for this
	elif ( part_code == 4 ) :
		body_part_string = "robe"
		# use images 2304-2880 for this
	elif ( part_code == 5 ) :
		body_part_string = "shieldarm"
		# use images 2880-3456 for this
	elif ( part_code == 6 ) :
		body_part_string = "gun1" 
		# use images 3456-4032 for this
	elif ( part_code == 7 ) :
		body_part_string = "feet"
		# use images 4032-4608 for this
	elif ( part_code == 8 ) :
		body_part_string = "boots1"
		# use images 4608-5184 for this
	elif ( part_code == 9 ) :
		body_part_string = "weaponarm"
		# use images 5184-5760 for this
	else :
		print "********************"
		print "*** END OF ULTIMATE TUX GUN MOTION RENDERING REACHED"
		print "********************"
		print " "
		print " Should be safe to terminate now... Terminating..."
		print "...trying 'sys.exit()' now..."
		# os.abort()
		# sys.exit()
		print "...obviously, that hasn't worked. "
		print " ... Termination seems not to work... using safe part name 'NOTHING'"
		print " "
		body_part_string = "NOTHING"
	print "BLENDER (set_current_part_string) : The part code calculated was: " , part_code

######################################################################
########################## 'MAIN PROGRAM' ############################
######################################################################

#--------------------
# If we do an ultimate rendering here, then we need to set the current
# part string.  Otherwise the part string set manually in the initialisation
# code above should also do the trick...
#
if ( do_ultimate_rendering == 1 ) :
	set_current_part_string()

#--------------------
# Now we check if maybe all the animation in the current direction has
# been properly recorded and if therefore it's time to rotate the camera 
# to a new position for a new run of the animation in a new direction.
#
if ( time % ( full_motion_cycle - 0 ) == 0 ) :
	do_cam_rot()
	prepare_scene_for_tux_part()
	do_rename_and_croppy()
else:
	prepare_scene_for_tux_part()
	do_rename_and_croppy()